(ns clojure-eval.editor
  (:require 
   ["@codemirror/language" :refer [foldGutter syntaxHighlighting defaultHighlightStyle]]
   ["@codemirror/history" :refer [history historyKeymap]]
   ["@codemirror/state" :refer [EditorState EditorSelection]]
   ["@codemirror/view" :as view :refer [EditorView]]
   [clojure-eval.sci :as sci :refer [!viewer]]
   [clojure.string :as str]
   [applied-science.js-interop :as j]
   [nextjournal.clojure-mode :as cm-clj :refer [clojureLanguage]]))

(def theme
  (.theme
   EditorView
   (j/lit {".cm-content" {:white-space "pre-wrap", :padding "10px 0"}
           "&.cm-focused" {:outline "none"}
           ".cm-line" {:padding "0 10px"
                       :line-height "1.6"
                       :font-size "16px"
                       :font-family "var(--code-font)"}
           ".cm-matchingBracket" {:border-bottom "2px solid var(--black)"
                                  :color "inherit"}
           ".cm-gutters" {:background "transparent", :border "none"}
           ".cm-gutterElement" {:margin-left "5px"}
           ;; only show cursor when focused
           ".cm-cursor" {:visibility "hidden"}
           "&.cm-focused .cm-cursor" {:visibility "visible"}})))

(defonce extensions
  #js
   [theme
    (history)
    (syntaxHighlighting defaultHighlightStyle)
    (view/drawSelection)
    (foldGutter)
    (.. EditorState -allowMultipleSelections (of true))
    cm-clj/default-extensions
    (.of view/keymap cm-clj/complete-keymap)
    (.of view/keymap historyKeymap)])

(defn make-state [extensions doc]
  (let [[doc ranges] (->> (re-seq #"\||<[^>]*?>|[^<>|]+" doc)
                          (reduce (fn [[^string doc ranges] match]
                                    (cond (= match "|")
                                          [(str doc) (conj ranges (.cursor EditorSelection (count doc)))]
                                          (str/starts-with? match "<")
                                          [(str doc (subs match 1 (dec (count match))))
                                           (conj ranges (.range EditorSelection (count doc)
                                                                (+ (count doc) (- (count match) 2))))]
                                          :else
                                          [(str doc match) ranges])) ["" []]))]
    (.create EditorState
             #js{:doc doc
                 :selection
                 (if (seq ranges)
                   (.create EditorSelection (to-array ranges))
                   js/undefined)
                 :extensions
                 (cond-> #js[(.. EditorState -allowMultipleSelections (of true))]
                   extensions
                   (j/push! extensions))})))

(defn inlineEval [source]
  (reset! !viewer
          (new EditorView
               (j/obj :state
                      (make-state
                       (.concat #js[extensions] #js[(sci/extension)])
                       source)
                      :parent (js/document.querySelector "#app")))))

(defn clojure []
  cm-clj/language-support)
